#!/usr/bin/perl -w
#
# SMpMS - Status Monitoring Platform for your MySQL ServerFarm
# # This does not means that you can play _SM_ with (the fscking) _Micro$oft_
#
# Author: BlueT <BlueT@BlueT.org>
# Since 2008/01
# 
# Modified:
# BlueT <BlueT@BlueT.org> : AnyEvent version: Since 2010/09

#~ use common::sense;
use FindBin qw($Bin);
use lib "$Bin/lib";
#~ use lib 'lib';
use feature 'say';
use AnyEvent;
use AnyEvent::DBI;
use Data::Dumper;
use Net::Time qw(inet_daytime);
use Class::Date qw(date);
use Date::Calc qw(Decode_Month);	# FIXME: this module has huge dependencies
use Config::JSON;	# FIXME: thhis module dependends on Moose, which's fat
use Time::HiRes;
use JSON::XS;	# for exporting .json

$| = 1;


my $debug = 1;
my $output_file = "./SMpMS.html";
my $json_file = "./SMpMS.json";
my $interval = 5;
my $page_refresh = 5;
my $db_timeout = 2;

my $config_file = "./SMpMS.conf";
my $function_lib_module = "SMpMS.pm";
my $dbuser = "smpms";
my $dbpasswd = "qwe123";
my $db_retry_max = 2;

my %mon_opt;
my %AllStatus;

print "\n\nStarting SMpMS - Status Monitoring Platform for your MySQL ServerFarm ".`date`."\n";

# read fom config
my %config = %{ &read_from_config($config_file) };

#~ # read functions from module
#~ open FUNC, $function_lib_module or die "cannot open module $function_lib_module: $!";
#~ my $func; while (<FUNC>) {chomp; s/(\s+)?#(.+)?$//gi; $func .= $_}; eval $func; close FUNC; undef $func;

# connect to DBs
my %DBs = %{ &initial_db_connection(\%config) };
#~ print Dumper \%DBs;


# FIXME: should reload config and re-initial all DB connections when $SIG{'HUP'}
$SIG{'HUP'} = sub {
	%config = %{ &read_from_config($config_file) };
	undef %DBs;
	%DBs = %{ &initial_db_connection(\%config) };
};


while (1) {
	#~ sleep $interval;	# FIXME: this should be configured in AE way.
	print `date`."\n";
	# fillout
	
	undef %DBs;
	%DBs = %{ &initial_db_connection(\%config) };
	
	#~ my ($db,$status) = &get_info_from_db(\%DBs, \%config);
	my ($status) = &get_info_from_db(\%DBs, \%config);
	#~ %DBs = %$db;
	%AllStatus = %$status;
	#~ print "\nAllStatus:\n";
	#~ print Dumper \%AllStatus;exit;
	
	# show/write final results to a file or screen.
	my %FinalResult = &compare(\%AllStatus, \%config);
	#~ print Dumper %FinalResult;exit;
	
	&output_json(\%FinalResult);
	&output_html(\%FinalResult);
	
	sleep $interval;	# FIXME: this should be configured in AE way.
	
	#exit;
}

#####

sub read_from_config {	# checked
	my $config_file = shift;
	my $conf;
	my %config;
	
	open CONF, $config_file or die "cannot open config file $config_file: $!";
	while (<CONF>) {chomp; s/(\s+)?#(.+)?//gi; $conf .= $_}; eval $conf; close CONF; $conf = "";
	
	print "conf loaded\n";
	
	return \%config;
}


sub initial_db_connection {	# AnyEvent::DBI ready
	my %config = %{ shift @_ };
	my %DBs;
	
	my $aecv_local = AE::cv;
	$aecv_local->begin (sub { shift->send });
	
	
	while ( my ($ip_dsc, $param) = each %config ) {
		my ($ip, $dsc) = split/-/,$ip_dsc;
		my ($port, $thread, $time, $io_running, $sql_running, $second_behined_master) = split/\s/,$param;
		
		my $dsn = "DBI:mysql:host=$ip;port=$port";
		print "$dsn\n";
		
		say "db_timeout: $db_timeout";
		
		if ($port) {
			$aecv_local->begin (sub { shift->send });
			my $dbh = new AnyEvent::DBI 
				$dsn, $dbuser, $dbpasswd,
				PrintError => 1,
				#~ timeout => $db_timeout,
				on_error   => sub {
			                #~ $log_handle->push_write ("DBI Error: $@ at $_[1]:$_[2]\n");
			                print "DBI Error: $@ at $_[1]:$_[2]\n";
					print "dbh: ".Dumper($DBs{"$ip:$port:$dsc"});
					#~ %DBs = %{ &initial_db_connection(\%config) };
					$aecv_local->end;
				},
				on_connect => sub {
					#~ my $dbh = shift;
					#~ say "set timeout" if $debug;
					#~ $dbh->timeout ($db_timeout);
					#~ 
					#~ say "set wait_timeout" if $debug;
					#~ $dbh->exec ("set wait_timeout=28800",,sub{});
					#~ 
					#~ say "set interactive_timeout" if $debug;
					#~ $dbh->exec ("set interactive_timeout=28800",,sub{});
					#~ 
					$aecv_local->end;
				};
			
			
			
			$DBs{"$ip:$port:$dsc"} = $dbh;
		} else {
			$DBs{"$ip:$port:$dsc"} = "";
		}
	}
	
	$aecv_local->end;
	
	$aecv_local->cb (sub {});
	$aecv_local->wait;
	#~ $aecv_local->recv;
	
	print "db inited\n";
	#~ print Dumper \%DBs;
	
	return \%DBs;
}


sub get_info_from_db {	# ready for async
	my %DBs = %{ shift @_ };
	my %config = %{ shift @_ };
	
	#~ print Dumper \%DBs if $debug;
	
	my $aecv = AE::cv;
	$aecv->begin (sub { shift->send (\%AllStatus) });
	
	my @need_chk_epoch;
	while ( my ($ip_dsc, $param) = each %config ) {
		
		my ($ip, $dsc) = split/-/,$ip_dsc;
		my ($port, $thread, $time, $io_running, $sql_running, $second_behined_master) = split/\s/,$param;
		
		my @monitoring_options;
		push @monitoring_options, \@{['Threads_connected', "show status like 'Threads_connected'"]} if $thread;
		push @monitoring_options, \@{['Slave_IO_Running', "show slave status"]} if $io_running;
		push @monitoring_options, \@{['Slave_SQL_Running', "show slave status"]} if $sql_running;
		push @monitoring_options, \@{['Seconds_Behind_Master', "show slave status"]} if $second_behined_master;
		$mon_opt{"$ip:$port:$dsc"} = \@monitoring_options;
		
		print "\n\nNEW START: $ip:$port\n";
		
		#~ $aecv->begin;
		#~ &async_check_db($aecv, "$ip:$port:$dsc");
		
		my $aecv_local = AE::cv;
		for my $pair ( @{$mon_opt{"$ip:$port:$dsc"}} ) {
			my ($item, $query) = @$pair;
			print "DSC_short: $dsc\n" . "QUERY: " . Dumper $query . "\nITEM: " . Dumper $item if $debug;
			&async_check_db($aecv, "$ip:$port:$dsc", $item, $query, $aecv_local);
		}
		$aecv_local->cb (sub {});
		$aecv_local->recv;
		
		push @chk_epoch_pool, "$ip:$port:$dsc" if $time;
		
	}
	
	# FIXME: re-confirm: do we still need NetTime?
	#~ &chk_epoch(@chk_epoch_pool);
	
	$aecv->end;
	say "Shoot All!";
	
	$aecv->cb (sub {});
	#~ $aecv->recv;
	$aecv->wait;
	
	print "AllStatus: " . Dumper \%AllStatus if $debug;
	
	return \%AllStatus;
}


sub async_check_db {	# ready for async
	my $aecv = shift;
	my $dsc = shift;
	my $item = shift;
	my $query = shift;
	my $aecv_local = shift;
	my $retry_count = shift || 0;
	say "Retry counter is $retry_count";
	if ($retry_count >= $db_retry_max) {
		say "Reached DB Retry Max";
		#~ say "GOING TO RE-INITIAL DB CONNECTIONS" if $debug;
		#~ undef %DBs;
		#~ %DBs = %{ &initial_db_connection(\%config) };
		#~ say "RE-INITIAL DB CONNECTIONS DONE" if $debug;
		$aecv->end;
		$aecv_local->end;
		say "SENT AECV END" if $debug;
		return;
	}
	
	print "DSC: $dsc\n" . "QUERY: " . Dumper $query . "\nITEM: " . Dumper $item if $debug;
	
	#~ print "MON_OPT: " . Dumper %mon_opt if $debug;
	
	#~ my $query = $mon_opt{$dsc}->[0][1];
	#~ for my $pair ( @{$mon_opt{$dsc}} ) {
		#~ say "PAIR: " . Dumper $pair;
		#~ my ($item, $query) = @$pair;
		#~ print "QUERY: " . Dumper $query . "\nITEM: " . Dumper $item if $debug;
		#~ print "DBs: " . Dumper %DBs if $debug;
		#~ print "$dsc: " . Dumper $DBs{"$dsc"} if $debug;
		#~ print "AECV: " . Dumper $aecv if $debug;
		
		#~ my $aecv_local = AE::cv;
		
		$aecv->begin (sub { shift->send });
		$aecv_local->begin (sub { shift->send });
		$DBs{"$dsc"}->exec_hash ($query, @args,
			sub {
				say "One more shoot." if $debug;
				#~ print "DSC: $dsc\n" . "QUERY: " . Dumper $query . "\nITEM: " . Dumper $item if $debug;
				my ($dbh, $rows, $rv) = @_;
				
				#~ my @row = @$rows;
				#~ $#_ or die "failure: $@";
				say "REF: ".ref $rows;
				print "ROWS: ";print Dumper $rows; print "\n";
				#~ print "ROW: ". Dumper $row[0]->[1] . "\n";
				
				#~ %result = @$_[1];
				#~ (my $item, undef) = @{shift @{$mon_opt{$dsc}} };
				#~ $AllStatus{"$ip:$port:$dsc"}{$item} = $result->{$item};
				#~ $AllStatus{"$dsc"}{$item} = $row[0]->[1];
				$AllStatus{"$dsc"}{$item} = $rows->{$item};
				#~ @$mon_opt{$dsc} > 0 ? &async_check_db($dsc,++$retry_count) : $aecv->end ;
				if ($#_) {
					$aecv->end;
					$aecv_local->end;
				} else {
					say "failure: $@";
					#~ %DBs = %{ &initial_db_connection(\%config) };
					&async_check_db($aecv, $dsc, $item, $query, $aecv_local, ++$retry_count)
				};
				
				#~ @$mon_opt{$dsc} > 0 ? &async_check_db($aecv, $dsc, ++$retry_count) : $aecv->broadcast;
				#~ say "AllStatus for this shot: " . Dumper %AllStatus;
			}
		);
	#~ }
	#~ $aecv_local->recv;
}



sub chk_epoch {	# FIXME: re-confirm: do we still need NetTime?
	for $dsc (@_) {
		my ($ip, $port, $sub_dsc) = split/:/,$dsc;
		(my $remote_epoch, $local_epoch) =  &get_epoch($ip,$local_epoch);
		
		$AllStatus{"$ip:$port:$dsc"}{'Time'} = $remote_epoch;
		$AllStatus{"$ip:$port:$dsc"}{'LocalTime'} =  $local_epoch = &get_epoch('localhost');
	}
}

sub get_epoch {	# FIXME: re-confirm: do we still need NetTime?
	my $ip = shift;
	my $local_epoch = shift;
	print "get_epoch\n";
	my $string;
	print "Epoch - orig: $local_epoch\n";
	
	my $epoch_error = 0;
	eval{
		local $SIG{'ALRM'}  = sub { $epoch_error = 1; return 0 };
		alarm($db_timeout);
		$string = inet_daytime($ip);
		alarm(0);
	};
	print "STRING: $string\n";
	if($@){
		print "Error: $@\n";
		print "daytime: ".Dumper($string);
		#~ exit;
	}
	
	print "EPOCH_ERROR is $epoch_error\n";
	($local_epoch, undef) = &get_epoch('localhost') if $epoch_error == 1;
	
	$string =~ s/[\n\r]//g;
	my @daytime = split/ /,$string;
	my $month = Decode_Month($daytime[1]);
	my $date = date "$daytime[2]-$month-$daytime[0] $daytime[3]", $daytime[4];
	
	#$date->set_time_zone( 'UTC' );
	
	print "DATESTRING: $string\n";
	print "DATE:  $daytime[2]-$month-$daytime[0] $daytime[3], and the Forth $daytime[4]\n";
	
	my $remote_epoch = scalar $date->epoch;
	print "DATE Epoch: $remote_epoch\n";
	
	# confirm, and re-confirm, to make sure $local_epoch exist
	if ($ip eq "localhost") {
		print "It's localhost!\n";
		while (not $remote_epoch =~ /^\d+$/) {
			print "Re-Try localhost : $remote_epoch\n";
			sleep 1;
			($remote_epoch, undef) = &get_epoch('localhost');
		}
	}
	
	print "This: $remote_epoch, Local: $local_epoch\n";
	
	wantarray? return ($remote_epoch, $local_epoch) : $remote_epoch;

}


sub compare {
	my %AllStatus = %{ shift@_ };
	my %config = %{ shift @_ };
	#~ my $local_epoch = &get_epoch('localhost');
	my %FinalResult;
	
	#~ print Dumper $AllStatus;# exit;
	#~ print Dumper $config; #exit;
	
	while ( my ($ip_dsc, $param) = each %config ) {
		my ($ip, $dsc) = split/-/,$ip_dsc;
		my ($port, $thread, $time, $io_running, $sql_running, $second_behined_master) = split/\s/,$param;
		my $warm;
		
		#~ 'Time' => 1201627208,
		my $diff_time = abs($AllStatus{"$ip:$port:$dsc"}{"Time"}-$AllStatus{"$ip:$port:$dsc"}{"LocalTime"} );
		$FinalResult{"$ip:$port:$dsc"}{"diff_time"} = ( $time and not $AllStatus{"$ip:$port:$dsc"}{"Time"} )  ? ( $warm=1 ) && ( "Alarm:X") : ( $time and $diff_time >= $time) ? ( $warm=1 ) && ( "Alarm:$diff_time") : $diff_time ;
		#~ print "epoch: $local_epoch, Time: ".$AllStatus{"$ip:$port"}{"Time"}.", diff: $diff_time, final: ".$FinalResult{"$ip:$port"}{"diff_time"}."\n";
		
		#~ 'Seconds_Behind_Master' => '0',
		my $sec_behind = $AllStatus{"$ip:$port:$dsc"}{"Seconds_Behind_Master"};
		$FinalResult{"$ip:$port:$dsc"}{"sec_behind"} = ( $second_behined_master and not defined $sec_behind )  ? ( $warm=1 ) && ( "Alarm:X") : ( $second_behined_master and $sec_behind >= $second_behined_master) ? ( $warm=1 ) && ( "Alarm:$sec_behind") : $sec_behind ;
		#~ print "\n$ip:$port - SB: $sec_behind, SBM: $second_behined_master, Final: ".$FinalResult{"$ip:$port"}{"sec_behind"}."\n";
		
		#~ 'threads' => '1'
		my $now_threads = $AllStatus{"$ip:$port:$dsc"}{"Threads_connected"};
		$FinalResult{"$ip:$port:$dsc"}{"now_threads"} = ( $thread and not defined $now_threads )  ? ( $warm=1 ) && ( "Alarm:X") : ( $thread and $now_threads >= $thread) ? ( $warm=1 ) && ( "Alarm:$now_threads") : $now_threads ; ;
		
		#~ 'IO_Running' => 'Yes',
		my $io = $AllStatus{"$ip:$port:$dsc"}{"Slave_IO_Running"};
		$FinalResult{"$ip:$port:$dsc"}{"IO"} = ( $io_running and not defined $io )  ? ( $warm=1 ) && ( "Alarm:X") : ( $io_running and not $io eq "Yes" ) ? ( $warm=1 ) && ( "Alarm:$io") : $io ;
		
		#~ 'SQL_Running' => 'Yes',
		my $sql = $AllStatus{"$ip:$port:$dsc"}{"Slave_SQL_Running"};
		$FinalResult{"$ip:$port:$dsc"}{"SQL"} = ( $sql_running and not defined $sql )  ? ( $warm=1 ) && ( "Alarm:X") : ( $sql_running and not $sql eq "Yes" ) ? ( $warm=1 ) && ( "Alarm:$sql") : $sql ;
		
		$FinalResult{"$ip:$port:$dsc"}{"Alarm"} = 1 if $warm;
		
	}
	
	#~ print Dumper %FinalResult; exit;
	return %FinalResult;	
}

sub output_json {
	my %FinalResult = %{ shift @_};
	my $tmp_file = $json_file.".tmp";
	
	print "\nGenerating Output json file...\n";
	open JSON, ">$tmp_file" or die "cannot open $tmp_file for tempolary write out! :$!";
	
	my $json = encode_json \%FinalResult;
	print JSON $json;
	
	close JSON;
	
	rename $tmp_file, $json_file or die "cannot rename $tmp_file to $json_file! :$!";
	print "\nOutput json file Generated.\n";
}

sub output_html {
	my %FinalResult = %{ shift @_};
	my $tmp_file = $output_file.".tmp";
	my $alarm;
	
	print "\nGenerating Output html file...\n";
	open HTML, ">$tmp_file" or die "cannot open $tmp_file for tempolary write out! :$!";
	
	print HTML "
		<html>
		<head>
		<title>SMpMS ::  Status Monitoring Platform for your MySQL ServerFarm</title>
		<meta content=\"text/html; charset=UTF-8\" http-equiv=\"content-type\"/>
		<META HTTP-EQUIV=\"Refresh\" CONTENT=\"$page_refresh\">
		</head>
		Last Sync Time: &nbsp; ".`date`."<br />
		<hr />
		<h3>DB ERROR in past $page_refresh seconds:</h3>
		<table border=\"1\" hight=\"100%\" width=\"100%\">
			<tbody>
	";
	
	
	# show if there's any DB ERROR
	for (sort {$a cmp $b} keys %FinalResult) {
		next if not $FinalResult{$_}{"Alarm"};
		my %Result = %{$FinalResult{$_}};
		delete $Result{"Alarm"};
		$alarm = 1;
		#~ print Dumper %Result;
		
		print HTML "
				<tr>
					<td>
						<table border='0' cellspacing='1' cellpadding='0' bgcolor='#808080' width='100%'>
							<tr height=\"50\"><td colspan='5' align='center' bgcolor=\"red\"><font size=\"45\" color=\"yellow\"><strong>$_</strong></font></td></tr>
							<tr style='font-size:11px' align='center' height='18' bgcolor='#C0C0C0'>
		";
		
		my $items;
		(print HTML "<td>$_</td>" and $items++) for sort {$a cmp $b} keys %Result;	# print Items
		my $item_width = 100/$items;
		
		print HTML "
							</tr>
							<tr height='25' align='center'>";
		
		for (sort {$a cmp $b} keys %Result) {	# print Values
			print HTML $Result{$_} =~ /Alarm/? ($Result{$_} =~ s/Alarm://) && ("<td bgcolor=\"red\" width='$item_width%'><font color=\"yellow\"><strong>$Result{$_}</strong></font></td>") : "<td width='$item_width%'>$Result{$_}</td>" ;
		}
		print HTML "
							</tr>
						</table>
					</td>
				</tr>\n
		";
		
	}
	
	print HTML '
			</tbody>
		</table>
		<hr />
		Author: BlueT@BlueT.org <br />
	';
	print HTML '<object><embed hidden="true" autostart="true" src="./webdown.mp3"/><param value="webdown.wav" name="FileName"/></object>' if $alarm ==1;
	print HTML "</html>";
	close HTML;
	
	rename $tmp_file, $output_file or die "cannot rename $tmp_file to $output_file! :$!";
	print "\nOutput html file Generated.\n";
}


__END__
